import click

from hookee.manager import Manager

from future.utils import iteritems

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.6"


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("--port", type=int, help="The port for the local webserver.")
@click.option("--plugins_dir", help="The directory to scan for custom `hookee` plugins.")
@click.option("--plugins", multiple=True, help="A list of `hookee` plugins to enable.")
@click.option("--auth_token", help="The `ngrok` auth token use.")
def hookee(ctx, **kwargs):
    """
    If options are given, they override the default values derived from the config file.
    """
    ctx.ensure_object(dict)
    for key, value in iteritems(kwargs):
        if value:
            ctx.obj[key] = value

    manager = Manager(ctx)

    if ctx.invoked_subcommand is None:
        manager.start()
    else:
        ctx.obj["manager"] = manager


@hookee.command()
@click.pass_context
def start(ctx):
    """
    Start `hookee`.
    """
    manager = ctx.obj["manager"]

    manager.start()


@hookee.command()
@click.pass_context
@click.argument("plugins_dig")
def set_plugins_dir(ctx, plugins_dir):
    """
    Set the default directory to be scanned for plugins.
    """
    manager = ctx.obj["manager"]

    manager.config.set("plugins_dir", plugins_dir)

    manager.print_util.print_config_update("Plugins dir has been updated.")


@hookee.command()
@click.pass_context
@click.argument("plugin")
def enable_plugin(ctx, plugin):
    """
    Enable the given plugin by default.
    """
    manager = ctx.obj["manager"]

    manager.validate_plugin(plugin)

    manager.config.append("plugins", plugin)

    manager.print_util.print_config_update("Plugin \"{}\" has been enabled.".format(plugin))


@hookee.command()
@click.pass_context
@click.argument("plugin")
def disable_plugin(ctx, plugin):
    """
    Disable the given plugin by default.
    """
    manager = ctx.obj["manager"]

    # TODO: because the manager is initialized in the main group, plugins are validated prior to this execute, but
    #   if trying to disable a plugin that is no longer valid, this will always fail

    manager.config.remove("plugins", plugin)

    manager.print_util.print_config_update("Plugin \"{}\" has been disabled.".format(plugin))


@hookee.command()
@click.pass_context
@click.argument("auth_token")
def set_auth_token(ctx, auth_token):
    """
    Set the default `ngrok` auth token.
    """
    manager = ctx.obj["manager"]

    manager.config.set("auth_token", auth_token)

    manager.print_util.print_config_update("Default auth token has been set.")


if __name__ == "__main__":
    hookee(obj={})



import click

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.6"

BLUEPRINT_PLUGIN = "blueprint"
REQUEST_PLUGIN = "request"
RESPONSE_PLUGIN = "response"

VALID_PLUGIN_TYPES = [BLUEPRINT_PLUGIN, REQUEST_PLUGIN, RESPONSE_PLUGIN]


class PrintUtil:
    def __init__(self, config):
        self.config = config

    @property
    def console_width(self):
        return self.config.get("console_width")

    def print_config_update(self, msg):
        click.secho("\n--> {}\n".format(msg), fg="green")

    def print_hookee_banner(self):
        click.echo("")
        click.secho("=" * self.console_width, fg="green", bold=True)
        click.secho("""                .__                   __                  
                |  |__   ____   ____ |  | __ ____   ____  
                |  |  \ /  _ \ /  _ \|  |/ // __ \_/ __ \ 
                |   Y  (  <_> |  <_> )    <\  ___/\  ___/ 
                |___|  /\____/ \____/|__|_ \\___  >\___  >
                     \/                   \/    \/     \/ 
                                                   v{}""".format(__version__), fg="green", bold=True)
        click.secho("=" * self.console_width, fg="green", bold=True)

    def print_open_header(self, title, fg="green"):
        width = int((self.console_width - len(title)) / 2)

        click.echo("")
        click.secho("{}{}{}".format("-" * width, title, "-" * width), fg=fg, bold=True)
        click.echo("")

    def print_close_header(self, delimiter="-", fg="green"):
        click.secho(delimiter * self.console_width, fg=fg, bold=True)


import os

import confuse

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.4"


class Config:
    auth_token = confuse.as_template(str)
    auth_token.default = None

    console_width = confuse.as_template(int)
    console_width.default = 80

    template = {
        "auth_token": auth_token,
        "port": int,
        "plugins_dir": confuse.Filename(),
        "plugins": list,
        "console_width": console_width
    }

    def __init__(self, ctx):
        self.ctx = ctx

        try:
            config = confuse.Configuration("hookee", __name__)
            config.set_args(ctx.obj)

            self.config_obj = config
            self.config_dir = self.config_obj.config_dir()
            self.config_filename = os.path.join(self.config_dir, confuse.CONFIG_FILENAME)

            self.config = config.get(self.template)

            plugins_dir = os.path.expanduser(self.config["plugins_dir"])
            if not os.path.exists(plugins_dir):
                os.makedirs(plugins_dir)
        except confuse.NotFoundError as e:
            ctx.fail("The config file is invalid: {}.".format(str(e)))
        except (confuse.ConfigReadError, ValueError):
            ctx.fail("The config file is not valid YAML.")

    def get(self, key):
        return self.config[key]

    def set(self, key, value):
        if value != self.config[key]:
            self.config[key] = value

            self.config_obj[key] = value
            with open(self.config_filename, "w") as f:
                f.write(self.config_obj.dump())

    def append(self, key, value):
        list_item = list(self.config[key])

        if value not in list_item:
            list_item.append(value)
            self.config[key] = list_item

            self.config_obj[key] = list_item
            with open(self.config_filename, "w") as f:
                f.write(self.config_obj.dump())

    def remove(self, key, value):
        list_item = list(self.config[key])

        if value in list_item:
            list_item.remove(value)
            self.config[key] = list_item

            self.config_obj[key] = list_item
            with open(self.config_filename, "w") as f:
                f.write(self.config_obj.dump())



import click
from flask import jsonify

from hookee import util

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.4"

plugin_type = util.RESPONSE_PLUGIN
manager = None


def setup(_manager):
    global manager

    manager = _manager


def run(request, response):
    if not response:
        response = jsonify({"Response": "Ok"})

    # TODO: pretty this up further
    click.secho("Status Code: {}".format(response.status_code), fg="magenta")
    if response.headers:
        click.secho("Headers: {}".format(dict(response.headers)), fg="magenta")
    if response.data:
        click.secho("Body: {}".format(response.data.decode("utf-8")), fg="magenta")

    return response



import threading
import time

import click

from pyngrok import ngrok
from pyngrok.conf import PyngrokConfig

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.4"


class Tunnel:
    def __init__(self, manager):
        self.manager = manager
        self.port = self.manager.config.get("port")

        self.pyngrok_config = PyngrokConfig(auth_token=self.manager.config.get("auth_token"))

        self.public_url = None
        self.ngrok_process = None
        self._thread = None

    def _loop(self):
        self._start_tunnel()

        thread = threading.current_thread()
        thread.alive = True
        while thread.alive:
            time.sleep(1)

        thread.alive = False

        self.stop()

    def start(self):
        if self._thread is None:
            self.manager.print_util.print_open_header("Opening Tunnel")

            self._thread = threading.Thread(target=self._loop)
            self._thread.daemon = True
            self._thread.start()

            while self.public_url is None:
                time.sleep(1)

            self.print_close_header()

    def _start_tunnel(self):
        self.public_url = ngrok.connect(self.port, pyngrok_config=self.pyngrok_config).replace("http", "https")
        self.ngrok_process = ngrok.get_ngrok_process()

    def stop(self):
        if self._thread:
            ngrok.kill()

            self.public_url = None
            self.ngrok_process = None
            self._thread = None

    def print_close_header(self):
        click.echo(
            "* Tunnel: {} -> http://127.0.0.1:{}".format(self.public_url.replace("http://", "https://"), self.port))
        click.echo("")
        self.manager.print_util.print_close_header()



from datetime import datetime

import click

from hookee import util

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.4"

plugin_type = util.REQUEST_PLUGIN
manager = None


def setup(_manager):
    global manager

    manager = _manager


def run(request):
    now = datetime.now()

    manager.print_util.print_close_header("=", fg="magenta")
    click.echo("")
    click.secho(
        "[{}] \"{} {} {}\"".format(now.isoformat(), request.method, request.base_url,
                                   request.environ["SERVER_PROTOCOL"]),
        fg="magenta")

    manager.print_util.print_open_header("Request", fg="magenta")

    return request



import click

from hookee import util

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.4"

plugin_type = util.REQUEST_PLUGIN
manager = None


def setup(_manager):
    global manager

    manager = _manager


def run(request):
    # TODO: pretty this up further
    if request.headers and "X-Forwarded-For" in request.headers:
        click.secho("Client IP: {}".format(request.headers.get("X-Forwarded-For")), fg="magenta")
    if request.headers:
        click.secho("Headers: {}".format(dict(request.headers)), fg="magenta")

    return request



import inspect
import os
import time

import click

from hookee import conf, util
from hookee.server import Server
from hookee.tunnel import Tunnel
from hookee.util import PrintUtil

from pluginbase import PluginBase

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.4"


class Manager:
    def __init__(self, ctx):
        self.ctx = ctx

        self.config = conf.Config(self.ctx)
        self.loaded_plugins = self.load_plugins()
        self.print_util = PrintUtil(self.config)

        # TODO: validate if no plugins are loaded (for instance, if no Blueprints loaded, hookee will hang on startup)

        self.tunnel = Tunnel(self)
        self.server = Server(self)

        self.alive = False

        self.print_util.print_hookee_banner()

    def start(self):
        if not self.alive:
            self.server.start()
            self.tunnel.start()
            self.alive = True

            self.print_ready()

            try:
                while self.alive:
                    time.sleep(1)
            except KeyboardInterrupt:
                pass

            self.stop()

    def stop(self):
        if self.alive:
            self.server.stop()
            self.tunnel._thread.alive = False

            # Wait for the other threads to teardown
            while self.server._thread and self.tunnel._thread:
                time.sleep(1)

            self.alive = False

    def print_ready(self):
        self.print_util.print_open_header("Registered Endpoints")

        rules = list(filter(lambda r: r.rule not in ["/shutdown", "/static/<path:filename>", "/status"],
                            self.server.app.url_map.iter_rules()))
        for rule in rules:
            click.secho("* {}{}".format(self.tunnel.public_url, rule.rule))
            click.secho("  Methods: {}".format(sorted(list(rule.methods))))
            click.echo("")

        self.print_util.print_close_header()
        click.echo("")
        click.secho("--> Ready, send a request to a registered endpoint ...", fg="green", bold=True)
        click.echo("")

    def validate_plugin(self, plugin_name):
        try:
            plugin = self.source.load_plugin(plugin_name)

            functions_list = [o[0] for o in inspect.getmembers(plugin, inspect.isfunction)]
            attributes = dir(plugin)

            if not all(elem in attributes for elem in ["plugin_type", "manager"]):
                self.ctx.fail("Plugin \"{}\" does not conform to the plugin spec.".format(plugin_name))

            if plugin.plugin_type in [util.REQUEST_PLUGIN, util.RESPONSE_PLUGIN]:
                all(elem in functions_list for elem in ["setup", "run"])
            elif plugin.plugin_type == util.BLUEPRINT_PLUGIN:
                all(elem in functions_list for elem in ["setup"])
            else:
                self.ctx.fail("Plugin \"{}\" is not a valid plugin type.".format(plugin_name))

            # TODO: additionally validate the functions have correct num args

            return plugin
        except ModuleNotFoundError:
            self.ctx.fail("Plugin \"{}\" could not be found.".format(plugin_name))

    def load_plugins(self):
        builtin_plugins_dir = os.path.normpath(os.path.join(os.path.abspath(os.path.dirname(__file__)), "plugins"))
        plugins_dir = self.config.get("plugins_dir")
        enabled_plugins = self.config.get("plugins")

        plugin_base = PluginBase(package="hookee.plugins",
                                 searchpath=[builtin_plugins_dir])
        self.source = plugin_base.make_plugin_source(searchpath=[plugins_dir])

        loaded_plugins = []
        for plugin_name in enabled_plugins:
            plugin = self.validate_plugin(plugin_name)
            plugin.setup(self)
            loaded_plugins.append(plugin)

        return loaded_plugins

    def get_plugins_by_type(self, plugin_type):
        return filter(lambda p: p.plugin_type == plugin_type, self.loaded_plugins)



import click

from flask import Blueprint, request

from hookee import util

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.4"

blueprint = Blueprint("default", __name__)

plugin_type = util.BLUEPRINT_PLUGIN
manager = None


def setup(_manager):
    global manager

    manager = _manager


@blueprint.route("/webhook",
                 methods=["GET", "HEAD", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "TRACE", "CONNECT"])
def webhook():
    for plugin in manager.get_plugins_by_type(util.REQUEST_PLUGIN):
        plugin.run(request)

    manager.print_util.print_open_header("Response", fg="magenta")

    response = None
    for plugin in manager.get_plugins_by_type(util.RESPONSE_PLUGIN):
        response = plugin.run(request, response)

    manager.print_util.print_close_header("=", fg="magenta")
    click.echo("")

    return response


@blueprint.route("/status")
def status():
    return "", 200


@blueprint.route("/shutdown", methods=["POST"])
def shutdown():
    request.environ.get("werkzeug.server.shutdown")()

    return "", 204
